"""
KeeDuplicate is a custom exception raised to indicate that a KeeNum
class received a duplicate entry for an enumeration.
"""
#  AGPL-3.0 license
#  Copyright (c) 2025 Asger Jon Vistisen
from __future__ import annotations

from typing import TYPE_CHECKING

from ...utilities import textFmt

if TYPE_CHECKING:  # pragma: no cover
  from typing import Any


class KeeDuplicate(Exception):
  """
  KeeDuplicate is a custom exception raised to indicate that a KeeNum
  class received a duplicate entry for an enumeration.
  """

  __slots__ = ('name', 'value')

  def __init__(self, name: str, value: Any) -> None:
    """Initialize the KeeDuplicate object."""
    self.name = name
    self.value = value
    Exception.__init__(self, )

  def __str__(self, ) -> str:
    """Return the string representation of the KeeDuplicate object."""
    infoSpec = """Duplicate KeeNum member '%s' with value '%s'!"""
    info = infoSpec % (self.name, str(self.value))
    return textFmt(info)

  __repr__ = __str__



"""
LabelBox subclasses AttriBox and changes the read-write behaviour to
read-many and write-once. If an instance of LabelBox receives a call to
instance get before any value has been set, it will attempt to create a
value like the AttriBox, but then this value counts as the one write
allowed. Alternatively, if the value is set before the first get,
the value set, subject to type guards, will henceforth be returned.

To indicate that the value must be set, the DELETED sentinel may be passed.

class Foo:
  bar = LabelBox[str](DELETED)  # require a value to be set before first get
  baz = LabelBox[str]('untitled')  # uses 'untitled' as default value
  #  If the default value is ever retrieved, it uses up the single write
  #  operation allowed.
"""
#  AGPL-3.0 license
#  Copyright (c) 2025 Asger Jon Vistisen
from __future__ import annotations

from typing import TYPE_CHECKING

from ..core.sentinels import DELETED
from ..waitaminute import WriteOnceError

from . import AttriBox

if TYPE_CHECKING:  # pragma: no cover
  from typing import Any


class LabelBox(AttriBox):
  """
  LabelBox subclasses AttriBox and changes the read-write behaviour to
  read-many and write-once. If an instance of LabelBox receives a call to
  instance get before any value has been set, it will attempt to create a
  value like the AttriBox, but then this value counts as the one write
  allowed. Alternatively, if the value is set before the first get,
  the value set, subject to type guards, will henceforth be returned.
  """

  def __instance_set__(self, value: Any, *args, **kwargs) -> None:
    """
    Sets the value of the field for the given instance. If the value is
    not set, it initializes it with a new instance of the field type.
    """
    instance = self.getContextInstance()
    fieldType = self.getFieldType()
    pvtName = self.getPrivateName()
    if value is DELETED:
      return setattr(instance, pvtName, DELETED)
    if hasattr(instance, pvtName):
      if getattr(instance, pvtName) is not DELETED:
        oldValue = getattr(instance, pvtName)
        raise WriteOnceError(self, oldValue, value)
    if isinstance(value, fieldType):
      return setattr(instance, pvtName, value)
    setattr(instance, pvtName, fieldType(value))



"""
KeeMember encapsulates a member of an enumeration.

Defining properties: (independent properties)
  - name: The name of the member. This is the name by which the member
  appears in the class body of the enumeration. It is the name passed to
  the '__set_name__' method. These must be unique with enumerations and
  must be uppercase. The case requirement is more than convention,
  it is enforced. When an enumeration class body contains key, value pairs
  with the key not in uppercase, it is understood to mean that the value
  does not represent a member of the enumeration.
  - value: The value of the member. Uniquely, this value is not required
  to be unique across members of the same enumeration.
  - index: The index of the member. This index specifies how many members
  are before this member in the enumeration.

KeeMember may be used directly or may be further subclassed. The KeeNum
classes are created by the KeeMeta class which define the class behaviour
post creation, and KeeMember define how members are included in the
enumeration.
"""
#  AGPL-3.0 license
#  Copyright (c) 2025 Asger Jon Vistisen
from __future__ import annotations

from typing import TYPE_CHECKING

from ..core import Object
from ..desc import Field
from ..utilities import maybe, textFmt
from ..waitaminute import VariableNotNone
from ..waitaminute.keenum import KeeCaseException

if TYPE_CHECKING:  # pragma: no cover
  from typing import Any, Self, Type, TypeAlias

  KEENUM: TypeAlias = Type[object]


class Kee(Object):
  """KeeMember encapsulates a member of an enumeration. """

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  NAMESPACE  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  #  Class Variables

  #  Fallback Variables

  #  Private Variables
  __field_index__ = None
  __field_type__ = None
  __field_value__ = None

  #  Public Variables
  name = Field()

  #  Virtual Variables
  type_ = Field()

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  GETTERS  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  def getValue(self) -> Any:
    """Subclasses are free to implement this method to modify how the
    underlying value is retrieved. By default, the first argument passed
    to the constructor is returned as the value of the member or if
    absent, then the name of the member is returned."""
    return maybe(self.__field_value__, self.name)

  @type_.GET
  def _getType(self) -> type:
    """
    This method must return the type of the object that would return
    from the 'getValue' method. The default implementation requires the
    type be defined in the brackets:

    FOO = Kee[int](69)  # The type is 'int'
    """
    return self.__field_type__

  @name.GET
  def _getName(self) -> str:
    """Set by 'Object.__set_name__' when the enumeration is created."""
    return self.__field_name__

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  SETTERS  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  #  The 'KeeSpace' namespace class expects to be able to set 'name' and
  #  'index' after the 'Kee' object is but before the enumeration class is
  #  created. Subclasses that change this behaviour must also implement
  #  these changes  in the 'KeeSpace' class.

  @name.SET
  def _setName(self, name: str) -> None:
    if not name.isupper():
      raise KeeCaseException(name)
    if self.__field_name__ is not None:
      raise VariableNotNone('name', self.__field_name__)
    self.__field_name__ = name

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  Python API   # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  @classmethod
  def __class_getitem__(cls, fieldType: type) -> Self:
    """
    This method is called when the class is used as a type hint.
    It allows the class to be used as a type hint for the value type of
    the enumeration member.
    """
    self = cls.__new__(cls)
    self.__field_type__ = fieldType
    return self

  def __call__(self, *args, **kwargs) -> Self:
    """
    The Kee class uses the special syntax similar to AttriBox from the
    'worktoy.desc' package: NAME = Kee[int](69). This does change the
    '__call__' method to perform the role of the '__init__' method,
    but with the caveat that '__call__' must return 'self'.
    """
    Object.__init__(self, *args, **kwargs)
    if len(args) == 1:
      if isinstance(args[0], self.type_):
        self.__field_value__ = args[0]
        return self
    self.__field_value__ = self.type_(*args, **kwargs)
    return self

  def __int__(self, ) -> int:
    return self.__field_index__

  def __str__(self) -> str:
    """Return the string representation of the member."""
    infoSpec = """<%s member: %s>"""
    clsName = type(self).__name__
    info = infoSpec % (clsName, self.name)
    return textFmt(info)

  __repr__ = __str__



"""
KeeSpaceHook collects the KeeMember objects encountered in the class
bodies of KeeNum classes. This happens during the 'setItemPhase'. To avoid
context leakage, the members are collected in the owning namespace object.
The namespace object is expected to implement a method called
'addEnumeration' which 'KeeHook' calls to register the members. The
'KeeHook' provides no further functionality than deciding what key,
value pairs to collect as future members of the enumeration.
"""
#  AGPL-3.0 license
#  Copyright (c) 2025 Asger Jon Vistisen
from __future__ import annotations

from typing import TYPE_CHECKING

from ..mcls.space_hooks import AbstractSpaceHook

from . import Kee

if TYPE_CHECKING:  # pragma: no cover
  from typing import Any, Type, TypeAlias, Callable
  from . import KeeMeta

  __INIT__: TypeAlias = Callable[[KeeMeta, Kee], None]
  __GET__: TypeAlias = Callable[[KeeMeta, KeeMeta, Type[KeeMeta]], Any]


class KeeSpaceHook(AbstractSpaceHook):
  """
  KeeSpaceHook collects the KeeMember objects encountered in the class
  bodies of KeeNum classes. This happens during the 'setItemPhase'. To avoid
  context leakage, the members are collected in the owning namespace object.
  The namespace object is expected to implement a method called
  'addEnumeration' which 'KeeHook' calls to register the members. The
  'KeeSpaceHook' provides no further functionality than deciding what key,
  value pairs to collect as future members of the enumeration.
  """

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  DOMAIN SPECIFIC  # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  def setItemPhase(self, key: str, val: Any, old: Any = None, ) -> bool:
    """Hook for setItem. This is called before the __setitem__ method of
    the namespace object is called. The default implementation does nothing
    and returns False. """
    if isinstance(val, Kee):
      self.space.addNum(key, val)
      return True
    return False



"""
Dispatcher encapsulates the mapping from type signature to function
objects and thus provides the core overloading functionality.
"""
#  AGPL-3.0 license
#  Copyright (c) 2025 Asger Jon Vistisen
from __future__ import annotations

import sys

from types import FunctionType as Func
from types import MethodType as Meth

from ..core import Object
from ..utilities import maybe, typeCast, perm
from ..waitaminute import TypeException, VariableNotNone
from ..waitaminute.desc import ReadOnlyError, ProtectedError
from ..waitaminute.dispatch import DispatchException

from . import TypeSig

from typing import TYPE_CHECKING

if TYPE_CHECKING:  # pragma: no cover
  from typing import Any, Callable, Never, TypeAlias, Optional, Self

  Method: TypeAlias = Callable[[Any, ...], Any]
  Decorator: TypeAlias = Callable[[Method], Self]
  SigFuncList: TypeAlias = list[tuple[TypeSig, Method]]
  SigFuncMap: TypeAlias = dict[TypeSig, Method]


class Dispatcher(Object):
  """
  Dispatcher encapsulates the mapping from type signature to function
  objects and thus provides the core overloading functionality.
  """

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  NAMESPACE  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  #  Private Variables
  __sig_funcs__ = None
  __fallback_func__ = None
  __field_name__ = None
  __field_owner__ = None
  __finalizer_func__ = None

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  GETTERS  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  def _getSigFuncList(self) -> SigFuncList:
    return maybe(self.__sig_funcs__, [])

  def _getSigFuncMap(self) -> SigFuncMap:
    return {sig: func for sig, func in self._getSigFuncList()}

  def _getFallbackFunction(self) -> Optional[Method]:
    return self.__fallback_func__

  def _getFinalizerFunction(self) -> Optional[Method]:
    """
    Get the finalizer function if it exists.
    """
    return self.__finalizer_func__

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  SETTERS  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  def addSigFunc(self, sig: TypeSig, func: Method) -> Method:
    """
    Add a signature-function pair to the internal signature-function map.
    """
    existing = self._getSigFuncList()
    self.__sig_funcs__ = [*existing, (sig, func,)]
    return func

  def setFallbackFunction(self, func: Method) -> Method:
    if not callable(func):
      raise TypeException('__fallback_func__', func, Func, Meth)
    if self.__fallback_func__ is not None:
      raise VariableNotNone('__fallback_func__', self.__fallback_func__)
    self.__fallback_func__ = func
    return func

  def setFinalizerFunction(self, func: Method) -> Method:
    """
    Set the finalizer function that will be called when the dispatcher is
    deleted or finalized.
    """
    if not callable(func):
      raise TypeException('__finalizer_func__', func, Func, Meth)
    if self.__finalizer_func__ is not None:
      raise VariableNotNone('__finalizer_func__', self.__finalizer_func__)
    self.__finalizer_func__ = func
    return func

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  Python API   # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  def __get__(self, instance: Any, owner: type) -> Any:
    """
    Descriptor protocol method to return a decorator that can be used to
    register functions with specific type signatures.
    """
    finalizer = self._getFinalizerFunction()
    if instance is None:
      return self
    sigFuncMap = self._getSigFuncMap()

    def dispatch(*args, **kwargs) -> Any:
      try:
        argSig = TypeSig.fromArgs(*args, )
        func = sigFuncMap.get(argSig, None)
        #  FASTEST
        if func is not None:
          return func(instance, *args, **kwargs)
        #  FAST
        for sig, func in sigFuncMap.items():
          if len(argSig) != len(sig):
            continue
          for arg, type_ in zip(args, sig):
            if not isinstance(arg, type_):
              break
          else:
            return func(instance, *args, **kwargs)
        #  SLOW
        for sig, func in sigFuncMap.items():
          castArgs = []
          if len(argSig) != len(sig):
            continue
          for arg, type_ in zip(args, sig):
            if isinstance(arg, type_):
              castArgs.append(arg)
              continue
            try:
              castedArg = typeCast(type_, arg)
            except (ValueError, TypeError):
              break
            else:
              castArgs.append(castedArg)
          else:
            return func(instance, *castArgs, **kwargs)
        #  FALLBACK
        fallback = self._getFallbackFunction()
        if callable(fallback):
          return fallback(instance, *args, **kwargs)
        raise DispatchException(self, args, )
      finally:
        _, exception, __ = sys.exc_info()
        if callable(finalizer):
          try:
            finalizer(instance, *args, **kwargs)
          except Exception as finalException:
            if exception is None:
              raise finalException
            raise exception from finalException

    return dispatch

  def __call__(self, instance: Any, *args, **kwargs) -> Any:
    """
    Allows invoking the dispatcher when accessed through the owning class.
    """
    return self.__get__(instance, type(instance))(*args, **kwargs)

  def __set__(self, instance: Any, value: Any, **kwargs) -> Never:
    """Illegal setter operation"""
    raise ReadOnlyError(self, instance, value)

  def __delete__(self, instance: Any, **kwargs) -> Never:
    """Illegal delete operation"""
    raise ProtectedError(instance, self, )

  def __set_name__(self, owner: type, name: str) -> None:
    self.__field_name__ = name
    self.__field_owner__ = owner
    self.swapAllTHIS(owner)

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  CONSTRUCTORS   # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  def clone(self, ) -> Self:
    """
    Create a clone of this DescriptorOverload instance.
    """
    newLoad = type(self)()
    newLoad.__sig_funcs__ = self._getSigFuncList()
    fallback = self._getFallbackFunction()
    if fallback is not None:
      newLoad.__fallback_func__ = fallback
    return newLoad

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  DOMAIN SPECIFIC  # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  def overload(self, *types: type) -> Decorator:
    def decorator(func: Method) -> Self:
      """Decorator to register the function with the given types."""
      self.addSigFunc(TypeSig(*types), func)
      return self

    return decorator

  def finalize(self, func: Method) -> Decorator:
    self.setFinalizerFunction(func)
    return self

  def fallback(self, func: Method) -> Decorator:
    self.setFallbackFunction(func)
    return self

  def flex(self, *types: type, ) -> Decorator:
    """
    Decorator to register a function that can handle any type signature.
    This function will be called if no other signature matches.
    """

    def decorator(func: Method) -> Self:
      for p in perm(*types, ):
        self.addSigFunc(TypeSig(*p), func)
      return self

    return decorator

  def swapAllTHIS(self, thisType: type) -> None:
    """
    Swap all occurrences of THIS in the registered signatures with the
    provided type.
    """
    for sig, func in self._getSigFuncList():
      sig.swapTHIS(thisType)



"""
The 'worktoy.keenum' module provides the enumerating KeeNum class.
"""
#  AGPL-3.0 license
#  Copyright (c) 2025 Asger Jon Vistisen
from __future__ import annotations

from ._kee_member import Kee
from ._kee_flag import KeeFlag
from ._kee_space_hook import KeeSpaceHook
from ._kee_flags_hook import KeeFlagsHook
from ._kee_space import KeeSpace
from ._kee_meta import KeeMeta
from ._kee_num import KeeNum
from ._kee_flags_space import KeeFlagsSpace
from ._kee_flags_meta import KeeFlagsMeta
from ._kee_flags import KeeFlags

__all__ = [
    'Kee',
    'KeeFlag',
    'KeeSpaceHook',
    'KeeSpace',
    'KeeMeta',
    'KeeNum',
    'KeeFlagsSpace',
    'KeeFlagsMeta',
    'KeeFlags',
]
